
from odoo import api, Command
from odoo.tests import tagged, common

from odoo.addons.loempia.util import sanitize_url
from odoo.addons.loempia.controllers.frontend import AppsFrontendController
from odoo.addons.loempia.tests.common import MockRequest, _init_loempia

import logging
import os
import shutil

logger = logging.getLogger(__name__)

@tagged('post_install', '-at_install')
class TestLoempiaRepository(common.TransactionCase):

    def setUp(self):
        super(TestLoempiaRepository, self).setUp()
        _init_loempia(self)
        self.website = self.env['website'].browse(1)

        self.maintainer = self.env.ref('loempia.user_tpl')
        self.manager = self.env.ref('base.user_admin')
        self.maintainer.write({'oauth_uid': "dummy_oauth_uid"})
        self.manager.write({'oauth_uid': "dummy_manager_oauth_uid"})
        self.portal = self.env.ref('base.demo_user0')

        self.Repositories = self.env["loempia.repo"]

        self.Controller = AppsFrontendController()

    def tearDown(self):
        for repo in self.Repositories:
            self.clean_repo(repo)
        super(TestLoempiaRepository, self).tearDown()

    def clean_repo(self, repo_id):
        if os.path.exists(repo_id.path):
            shutil.rmtree(repo_id.path)
        if os.path.exists(repo_id.snap_path):
            shutil.rmtree(repo_id.snap_path)

    def _create_repo(self, url='ssh://git@github.com/ndeodoo/essai.git#13.0'):
        series_id = self.env['loempia.series'].get_default()
        if len(url.split('#')) > 1:
            series_id = self.env['loempia.series'].search([('name', '=', url.split('#')[1])]) or self.env['loempia.series'].get_default()
        return self.env['loempia.repo'].create({
            'series_id': series_id.id,
            'url': url,
            'vcs': 'git',
            'maintainer_id': self.maintainer.id,
            'state': 'active',
        })

    def _create_module(self, module_name, repo_id, **params):
        template = params.get('template_id') or self.env['loempia.module.template'].search([('name', '=', module_name), ('maintainer_id', '=', self.maintainer.id)])
        return self.env['loempia.module'].create({
            'name': module_name,
            'version': '1.0',
            'license': 'LGPL-3',
            'currency_id': self.env.ref('base.EUR').id,
            'price_single': 10.0,
            'repo_id': repo_id.id,
            'status': 'approved',
            'template_id': template.id if template else self._create_template(module_name).id,
            'depend_ids': [] if not params.get('depend_ids') else params.get('depend_ids')
        })

    def _create_template(self, module_name):
        return self.env['loempia.module.template'].create(
            {
                'name': module_name,
                'maintainer_id': self.maintainer.id,
            })

    def test_sanitize_url(self):
        # TODO test .git at the end when sanitized
        urls = [
            "git@github.com:ndeodoo/essai.git#10.0",
            "https://github.com/ndeodoo/essai.git#10.0",
            "https://dumb:dumb@github.com/ndeodoo/essai.git#10.0",
            "ssh://dumb:dumb@github.com/ndeodoo/essai.git#10.0"
        ]
        sanitized_urls = [
            "ssh://git@github.com/ndeodoo/essai.git#10.0",
            "ssh://git@github.com/ndeodoo/essai.git#10.0",
            "ssh://git@github.com/ndeodoo/essai.git#10.0",
            "ssh://dumb@github.com/ndeodoo/essai.git#10.0"
        ]
        for i in range(len(urls)):
            url = sanitize_url(urls[i])
            self.assertEqual(url, sanitized_urls[i], msg="Sanitize url repo")

    def test_01_check_url(self):
        """
        This test ensure that "_check_url" return True if the provided url
        is valid.
        """
        repo = self._create_repo()  # Default url
        self.assertTrue(repo._check_url(), msg="Correct url")
        self.Repositories |= repo

    def test_02_check_url(self):
        """
        This test ensure that "_check_url" log an error message in scan log if there
        is a space in the url.
        """

        error_msg = "The repository URL cannot contain spaces."
        with self.assertLogs('odoo.addons.loempia.util', level='WARNING') as log_catcher:
            repo = self._create_repo(url="ssh://git@github.com/ndeodoo essai.git#10.0")
            self.assertEqual(
                log_catcher.output[0],
                'WARNING:odoo.addons.loempia.util:Warning: ' + error_msg
            )
        self.assertTrue(error_msg in repo.scan_log, msg="Spaces in repo should log an error in scan log")
        self.Repositories |= repo

    def test_03_check_url(self):
        """
        This test ensure that "_check_url" log an error message in scan log if the
        version is missing in the url.
        """
        error_msg = "The repository URL must contain a branch matching exactly the series name."  # Only check begining of the error
        with self.assertLogs('odoo.addons.loempia.util', level='WARNING') as log_catcher:
            repo = self._create_repo(url="ssh://git@github.com/ndeodoo/essai.git")
            log_msg = "WARNING:odoo.addons.loempia.util:Warning: " + error_msg
            self.assertTrue(log_catcher.output[0].startswith(log_msg))
        self.assertTrue(error_msg in repo.scan_log, msg="Version should be present in repo url")
        self.Repositories |= repo

    def test_url_symlink(self):
        clean_path = self.env['loempia.repo']._url_symlink("/tmp", "ssh://git@github.com/ndeodoo essai.git#13.0")
        correct_value = "/tmp/ssh___git_github.com_ndeodoo_essai.git_13.0"
        self.assertEqual(correct_value, clean_path, msg="special chars should be removed")

    def test_path(self):
        repo = self._create_repo()
        path = "/tmp/repo_cache/ssh___git_github.com_ndeodoo_essai.git_13.0/repo"
        self.assertEqual(repo.path, path)
        self.assertTrue(os.path.exists(os.path.dirname(repo.path)), msg="It should create the repo")
        self.Repositories |= repo

    def test_snap_path(self):
        repo = self._create_repo()
        path = "/tmp/repo_cache/ssh___git_github.com_ndeodoo_essai.git_13.0/snapshot"
        self.assertEqual(repo.snap_path, path)
        self.assertTrue(os.path.exists(repo.snap_path), msg="It should create the repo")
        self.Repositories |= repo

    def test_is_server(self):
        repo = self._create_repo(url="ssh://git@github.com/odoo/odoo.git#13.0")
        self.assertTrue(repo.is_server, msg="Server repo")

        repo_bis = self._create_repo()
        self.assertTrue(not repo_bis.is_server, msg="Not a server repo")
        self.Repositories |= repo
        self.Repositories |= repo_bis

    def test_is_odoo_enterprise(self):
        repo = self._create_repo(url="ssh://git@github.com/odoo/enterprise.git#13.0")
        self.assertTrue(repo.is_odoo_enterprise, msg="Odoo enterprise repo")

        repo_bis = self._create_repo()
        self.assertTrue(not repo_bis.is_odoo_enterprise, msg="Not an odoo enterprise repo")
        self.Repositories |= repo
        self.Repositories |= repo_bis

    def test_01_is_maintainer(self):
        repo = self._create_repo()
        self.assertTrue(repo.with_user(self.manager).is_maintainer, msg="Admin user is maintainer of all repos")
        self.Repositories |= repo

    def test_02_is_maintainer(self):
        repo = self._create_repo()
        self.assertTrue(repo.with_user(self.maintainer).is_maintainer, msg="Maintainer of the repo")
        self.Repositories |= repo

    def test_03_is_maintainer(self):
        repo = self._create_repo()
        self.assertTrue(not repo.with_user(self.portal).is_maintainer, msg="Portal user is not maintainer of this repo")
        self.Repositories |= repo

    def test_publish_repo(self):
        repo = self._create_repo()
        module_names = ["A", "B"]
        modules = self.env['loempia.module']
        for name in module_names:
            modules |= self._create_module(name, repo)
        repo._publish_repo()
        self.assertTrue(all(modules.mapped("is_published")), msg="all modules should be published")

        self.Repositories |= repo

    def test_unpublish_repo(self):
        repo = self._create_repo()
        module_names = ["A", "B"]
        modules = self.env['loempia.module']
        for name in module_names:
            modules |= self._create_module(name, repo)
        repo._unpublish_repo()
        self.assertTrue(not any(modules.mapped("is_published")), msg="all modules should be unpublished")

        self.Repositories |= repo

    def test_module_invalidation(self):
        repo = self._create_repo()
        initial_status = ['approved', 'invalid', 'banned', 'blocked']
        final_status = ['invalid', 'invalid', 'banned', 'blocked']
        module_A = self._create_module("A", repo)

        # Ensure name filtering is correct
        repo._invalidate_modules(['B'])
        self.assertTrue(module_A.status == 'approved', msg="Invalidation of module B should not invalidate A")

        # Ensure that invalidation take care of previous status
        obtained_status = []
        for status in initial_status:
            module_A.status = status
            repo._invalidate_modules(['A'])
            obtained_status.append(module_A.status)
        self.assertTrue(obtained_status == final_status, msg="Invalidation should take care of the current status of the module")

    def test_01_upload_repo(self):
        env = api.Environment(self.env.cr, self.maintainer.id, {})
        url = "https://github.com/ndeodoo/essai.git#Fake"
        with MockRequest(env, website=self.website.with_env(env)):
            result = self.Controller.upload_repo(url, repo_id=None).get('error')
            self.assertEqual(
                "Your branch must match exactly the name of the series your module is meant for. Ex: #8.0 for 8.0 series.",
                result,
                msg="Error obtained when missing branch in url"
            )

    def test_02_upload_repo(self):
        repo = self._create_repo()
        env = api.Environment(self.env.cr, self.maintainer.id, {})
        url = "https://github.com/ndeodoo/essai.git#13.0"
        with MockRequest(env, website=self.website.with_env(env)):
            result = self.Controller.upload_repo(url, repo_id=None).get('error')
            self.assertEqual(
                "This repository and branch are already registered.",
                result,
                msg="Url of repositories must be unique"
            )
        self.Repositories = repo

    def test_03_upload_repo(self):
        repo = self._create_repo()
        repo_bis = self._create_repo(url="https://github.com/ndeodoo/essai2.git#13.0")
        env = api.Environment(self.env.cr, self.maintainer.id, {})
        url = "https://github.com/ndeodoo/essai.git#13.0"
        with MockRequest(env, website=self.website.with_env(env)):
            result = self.Controller.upload_repo(url, repo_id=repo_bis.id).get('error')
            self.assertEqual(
                "This repository and branch are already registered.",
                result,
                msg="Can not set the url of this repositories to the url of an exisiting repositories"
            )

        self.Repositories |= repo
        self.Repositories |= repo_bis

    def test_04_upload_repo(self):
        repo = self._create_repo()
        repo.state = 'active'
        _ = self._create_module("A", repo)
        env = api.Environment(self.env.cr, self.maintainer.id, {})
        url = "https://github.com/ndeodoo/essai.git#12.0"
        with MockRequest(env, website=self.website.with_env(env)):
            result = self.Controller.upload_repo(url, repo_id=repo.id).get('error')
            self.assertEqual(
                "You can not change the branch of an active repository. Please create a new one.",
                result,
                msg="A branch of an active repository (containing modules) cannot be changed"
            )

        self.Repositories |= repo

    def test_unlink_repo(self):
        """
        This test ensure that unlink a repository will clean all path, snapshot and so on.
        However, it is impossible to test it with modules as it requires to contact the
        account database.
        """
        repo = self._create_repo()
        path = repo.path
        snap_path = repo.snap_path
        repo.unlink()
        self.assertTrue(not os.path.exists(path), msg="Directories should be clean")
        self.assertTrue(not os.path.exists(snap_path), msg="Directories should be clean")

    def test_unlink_repo_with_modules_interdependencies(self):
        repo = self._create_repo()
        mod1 = self._create_module("mod1", repo)
        mod2 = self._create_module("mod2", repo, depend_ids=[Command.link(mod1.id)])
        mod3 = self._create_module("mod3", repo, depend_ids=[Command.link(mod1.id)])
        mod4 = self._create_module("mod4", repo)
        mod2.write({'depend_ids': [Command.link(mod4.id)]})

        repo.unlink()  # should not raise
        self.assertFalse(repo.exists())
        self.assertFalse((mod1 | mod2 | mod3 | mod4).exists())

    def test_01_dashboard_repo(self):
        """
        Basic test to ensure that dashboard filter works correctly.
        """
        series = [13.0, 12.0, 11.0, 10.0]
        for elem in series:
            url = "ssh://git@github.com/ndeodoo/essai.git#%s" % str(elem)
            self.Repositories |= self._create_repo(url=url)

        sorted_urls = self.Repositories.sorted('url').mapped('url')
        env = api.Environment(self.env.cr, self.maintainer.id, {})
        with MockRequest(env, website=self.website.with_env(env)):
            response = self.Controller.dashboard_repos(search=None, search_in='all')
            self.assertEqual(response.qcontext.get('repo_ids').sorted('url').mapped('url'), sorted_urls)

            response = self.Controller.dashboard_repos(search="ssh://git@github.com/ndeodoo/essai.git#13.0", search_in='url')
            self.assertEqual(response.qcontext.get('repo_ids')[0].url, "ssh://git@github.com/ndeodoo/essai.git#13.0")

            response = self.Controller.dashboard_repos(search="13.0", search_in='series')
            self.assertEqual(response.qcontext.get('repo_ids')[0].url, "ssh://git@github.com/ndeodoo/essai.git#13.0")

    def test_02_dashboard_repo(self):
        """
        Basic test to ensure that user_id is effectively added in the domain when redering
        the dashboard repos page.
        """
        series = [13.0, 12.0, 11.0, 10.0]
        for elem in series:
            url = "ssh://git@github.com/ndeodoo/essai.git#%s" % str(elem)
            self.Repositories |= self._create_repo(url=url)

        env = api.Environment(self.env.cr, self.manager.id, {})
        with MockRequest(env, website=self.website.with_env(env)):
            response = self.Controller.dashboard_repos(search=None, search_in='all')
            self.assertTrue(not response.qcontext.get('repo_ids'))
